// Термосервер на 4 датчика, 3-х сегментный индикатор. Ver.43 20.10.17 ZolSN & JackB 
// Написан с использованием стандартных примеров работы с термодатчиками серии DS18
// Работает с четырмя типами датчиков DS18
// Используются готовые библиотеки и примеры из сети
////
// Библия термо датчиков
#include <OneWire.h>
// Библия I2C
#include <Wire.h>
#define SLAVE_ADDRESS       0x07         // Адрес шины i2c для связи с внешним миром. Slave
//---------------------------------------------------------------------------------------
// Код индикатора позаимствован у OldBean 
//Индикатор
#define COMMON_CATHODE                  // Если индикатор с общим анодом - необходимо закомментировать эту строку
//---------------------------------------------------------------------------------------
#ifdef COMMON_CATHODE                   // Для индикаторов с общим катодом
#define L_HIGH 1
#define L_LOW  0
#else                                   // Для индикаторов с общим анодом
#define L_HIGH 0
#define L_LOW  1
#endif
//---------------------------------------------------------------------------------------
// Пины, к которым подключены сегменты индикатора - аноды (или катоды для индикаторов с общими анодами)
static uint8_t segPins[8] = {12, 16, 5, 7, 8, 13, 4, 6};// A B C D E F G Dp
static int d[3];                                        // Сотни, десятки и единицы показываемые на разрядах 7-сегментного индикатора 
// Пины, к которым подключены общие катоды (или аноды, для индикаторов с общими анодами)
static uint8_t digPins[3] = {15, 14, 11};
//                             gfedcba             
static uint8_t digit[11] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000}; // "Цифра" 10 это пробел
static uint8_t Dp = 0b10000000;          // Десятичная точка                   
static uint8_t mask = 0x01;              // Маска для анализа цифр
static int seg = 3;                      // Номер сегмента, обрабатываемого (подствечиваемого, если нужно) в данный момент
static volatile int cntr = 0;            // Счетчик для тактирования смены подствечиваемых сегментов
//---------------------------------------------------------------------------------------
// Код кнопки переписан на основе примеров
// Кнопка
#define NO_ACT              0
#define BOTTON_CLICK        0b00000010  // Короткое нажатие кнопки 
#define BOTTON_PRESS        0b00001000  // Удержание кнопки  в нажатом состоянии
static uint8_t mButPin = 17;            // Пин, к которому подключена кнопка 
static int dtBounce = 10;               // Задержка (мс) для устранения дребезга контактов кнопок
static int dtPress = 300;               // Задержка, после которой считается, что кнопка удерживается
static uint8_t bos = NO_ACT;            // Предыдущее состояние кнопок
static unsigned long tpr;               // Момент нажатия кнопки
bool  button_state      = false;        // Статус кнопки
bool  button_long_state = false;        // статус длинного нажатия
//---------------------------------------------------------------------------------------
static int indx = 1;                   // текущий индекс датчика
// Первый байт содержит маску присутствия датчика и статус готовности данных
// 4 бита на статус датчика 1-есть датчик, 0-отсутствует. 
// 5-й бит идет замер температуры
// далее по два байта значение темпратуры на каждый датчик
byte byteTemp[9];                      // статусы и замеры
// Датчики мксимальное кол-во
int MaxDs=4;
bool Ready_Ds_Data = false;
// на каких пинах сидят датчики
// задействованы пины 0,1,2,3
OneWire ds[4]={OneWire(0),OneWire(1),OneWire(2),OneWire(3)};

// Инициализация 
void setup(void) {  
// без этой команды библия датчиков не работает, почему не разбирался.
 Serial.begin(9600);
 
//Регистрирурем адрес I2C
 Wire.begin(SLAVE_ADDRESS); // Инициализация библиотеки для шины i2c
 Wire.onReceive(receiveData); 
 Wire.onRequest(sendData);
 
// Инициализация пинов, к которым подключены сегменты индикатора
  for(int i = 0; i < 8; i++) { 
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], L_LOW);
  }
// Инициализация пинов, к которым подключены общие катоды (аноды) индикатора  
  for(int i = 0; i < 3; i++) { 
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], L_HIGH);
  }
// незнаю почему, но остался поднятым и мешал работе  
 digitalWrite(16, L_LOW);         
  
// Инициализация таймера 2 для тактирования системы
  TCCR2A = (1<<WGM21);            // Режим CTC (сброс по совпадению)
  TCCR2B = (1<<CS21) | (1<<CS20); // Предделитель 32
  OCR2A = 98;                     // При этом значении совпадения происходят с частотой около 5 кГц.
  TIMSK2 = (1<<OCIE2A);           // Разрешить прерывание по совпадению

// Если используются пины 0 и 1 для датчиков, то отрубаем RxTx
  UCSR0B &= ~(1<<TXEN0);  // Гасим Tx
  UCSR0B &= ~(1<<RXEN0);  // Гасим Rx
  pinMode(1, INPUT);      // Перевод на вход Tx
  pinMode(0, INPUT);      // Перевод на вход Rx
  digitalWrite(0, L_HIGH); // Подпор 5V на Rx  
}

//---------------------------------------------------------------------------------------
// смотрим за кнопкой
uint8_t GetKey(void) {  
    bool pin_state = digitalRead(mButPin);
   // Фиксируем нажатие кнопки   
   if( pin_state  == LOW && !button_state  ){
      button_state      = true;
      button_long_state = false;  
      tpr = millis(); 
   }
   // Фиксируем длинное нажатие кнопки   
   if( pin_state  == LOW && !button_long_state && ( millis() - tpr ) > dtPress ){
      button_long_state = true;
      return BOTTON_PRESS;      
   }
// Фиксируем отпускание кнопки   
   if( pin_state == HIGH && button_state && ( millis() - tpr ) > dtBounce  ){
      button_state      = false;   
      tpr       = millis();
      if( !button_long_state ) return BOTTON_CLICK;      
   }   
   return NO_ACT;
}
//---------------------------------------------------------------------------------------
// таймер прерывания для индикации и отлавливает кнопку
ISR (TIMER2_COMPA_vect) {  
  uint8_t ks = GetKey();         
  switch(ks) {
    case BOTTON_CLICK:  indx++;                    // переход к следующему датчику
                      if(indx > MaxDs) indx = 1;                                                                                                           
                      break;    
    case BOTTON_PRESS:  d[2]=10;d[0]=10;d[1]=indx; //пока держим переопределим индикаторы.                                                
                      break;    
                      }
                      
// поджигаем индикатор                      
   if(++cntr >= 10)
   { // Каждые 2 мс переходим к посветке следующей группы сегментов    
    cntr = 0;        
    ShowSegments();    
   }       
}

// Подготовка индикаторов (цифр) для визуализации 
void UpdateDigits(int val) { 
  d[2] = val/100;
  d[1] = (val - d[2]*100)/10;
  d[0] = val - d[2]*100 - d[1]*10;
  if(d[2] == 0) 
    d[2] = 10;        
}
//---------------------------------------------------------------------------------------
// Комментарии написаны для индикаторов с общими катодами. В скобках - для индикаторов с общими анодами
void ShowSegments() {                 // Подсвечивание очередной группы сегментов                                      
  digitalWrite(segPins[seg], L_LOW);  // Опускаем (поднимаем) потенциал анода (катода) "отработавшего" сегмента  
  if(++seg < 8) mask <<= 1;           // Переходим к следующему сегменту
  else {                              // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }    
  digitalWrite(segPins[seg], L_HIGH); // Поднимаем (роняем) потенциал анода (катода) нового текущего сегмента
  for(int bi = 0; bi < 3; bi++) {     // Теперь пробегаем по цифрам индикатора  
    uint8_t digitDp=digit[d[bi]];
    if(bi==1) digitDp|=Dp;            //поджикаем точку во втором разраде
    if(digitDp & mask)                // Если данный сегмент в bi-ом разряде должен гореть,
      digitalWrite(digPins[bi], L_LOW);  // то опустим (поднимем) потенциал катода (анода) соответствующей цифры
    else                                 // А если гореть не должен, то
      digitalWrite(digPins[bi], L_HIGH); // потенциал общего катода (анода) bi-й цифры должен быть высокий (низкий)
  }
}

//Главный цикл программа, его необходимо модернизировать для оптимальной работы. Надо будет перевести в прерывание
void loop(void) {  
  byte present = 0; 
  
  byte addr[MaxDs][8];          // адреса датчиков
  byte type_s[MaxDs];           // типы датчиков
  byte data[MaxDs][12];         // замеры датчиков сырые данные

  
  uint8_t inDs=0x01;
  //обходим датчики и шлем комаду на измерение
  for(int i=0;i<MaxDs;i++)
  {  
    //получаем адрес датчика     
    if ( !ds[i].search(addr[i])) { ds[i].reset_search(); }
    else  
    {
      //по адресу определяем тип используемого датчика.
      // если тип датчика определен маскируем первый байт наской присутствия датчика
      switch (addr[i][0]) {
        case 0x10:   //  Serial.println("  Chip = DS18S20");  // or old DS1820
           type_s[i] = 1; byteTemp[0]|=inDs; break;
        case 0x28: // Serial.println("  Chip = DS18B20");
           type_s[i] = 0; byteTemp[0]|=inDs; break;
        case 0x22:  //Serial.println("  Chip = DS1822");
           type_s[i] = 0; byteTemp[0]|=inDs;  break;
        default:    // тип не известен или сдох 
           type_s[i] = -1;  //Serial.println("  NoChip = DS"); break;  
      }
      // тип определен отправляем запрос на измерение
      if(type_s[i]>=0)
      {    
        ds[i].reset();  ds[i].select(addr[i]); ds[i].write(0x44, 1);            
      }        
    }
    inDs<<=1;  //сдвиг маски для следующего датчика   
  }
  //Ready_Ds_Data = false;  
  delay(850);     // ждем.  0,75сек. идет замер датчикм температуры
  //Данные еще не готовы поэтому если в запрос придет в этот момент передатим 5-й бит: Данные не готовы
  Ready_Ds_Data = false;
  
  // температура измерена, переведем датчики на отправку температуры
  for(int i=0;i<MaxDs;i++)
  { 
    // тип датчика известен, получим данные   
    if(type_s[i]>=0) {
      present = ds[i].reset();    
      ds[i].select(addr[i]);    
      ds[i].write(0xBE,1);         
      // читаем температуру в байтах
      for (int bi = 0; bi < 9; bi++)               
        data[i][bi] = ds[i].read();      
        
      //Переводим в сырой вид      
      int16_t raw = (data[i][1] << 8) | data[i][0];
      //определим поправку в зависимости от типа датчика
      if (type_s[i]) {
        raw = raw << 3; // 9 bit resolution default
        if (data[i][7] == 0x10) {   // "count remain" gives full 12 bit resolution
           raw = (raw & 0xFFF0) + 12 - data[i][6];
        }
      } else {
         byte cfg = (data[i][4] & 0x60);  // at lower res, the low bits are undefined, so let's zero them
         if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
         else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
         else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
         //// default is 12 bit resolution, 750 ms conversion time
      }

      //перевод в градусы цельсия
      float celsius = (float)raw / 16.0; // готовая температура датчика i
      // Для отправки по I2C необходимо целое число.
      int celsius100=(int)(celsius*100);
      // разобьем замеренную температуру на два байта для отправки по I2C
      byteTemp[1+2*i]=highByte(celsius100);
      byteTemp[2+2*i]=lowByte(celsius100);
      
      //для текущего(выбранного) датчика подкотовим отображение
      if((i+1)==indx)
      {
        //Индикатор 3-х разрабный поэтому только один десятичный знак
         UpdateDigits((int)(celsius*10));         
      }
    }
  }
  // данные в буфере имеются, можно смело отдавать по I2C
  Ready_Ds_Data = true;
  // возвращаемся измерять температуру следуюшим циклом.
  
  
}
// функции обратного вызова I2C
//---------------------------------------------------------------------------------------
// заглушка тупо читаем что пришло.
void receiveData(int byteCount) { 
  while(Wire.available()) {
     Wire.read();
  }
}
//---------------------------------------------------------------------------------------
// Отправка готовых данных.
void sendData() { 
  
  if(Ready_Ds_Data) // Данные готовы, отправим все
    Wire.write(byteTemp,9);
   else             // Данных в буфере еще нет, отправим только 5-й бит
    Wire.write((byte)16);
}

